import numpy as np
import cv2
import matplotlib.pyplot as plt

def generate_straight(length, width, start, direction=(1, 0)):
    direction = np.array(direction)
    norm_dir = direction / np.linalg.norm(direction)
    left_edge = [start + norm_dir * width / 2 * np.array([-1, 1]) * i for i in range(length)]
    right_edge = [start + norm_dir * width / 2 * np.array([1, -1]) * i for i in range(length)]
    return left_edge, right_edge

def generate_hairpin(radius, width, center, entry_angle=0):
    angles = np.linspace(entry_angle, entry_angle + np.pi, 180)
    left_edge = [center + (radius + width / 2) * np.array([np.cos(a), np.sin(a)]) for a in angles]
    right_edge = [center + (radius - width / 2) * np.array([np.cos(a), np.sin(a)]) for a in angles]
    return left_edge, right_edge

def generate_chicane(segment_lengths, width, start, direction=(1, 0)):
    direction = np.array(direction)
    norm_dir = direction / np.linalg.norm(direction)
    orth_dir = np.array([-norm_dir[1], norm_dir[0]])
    
    points = [start]
    for i, length in enumerate(segment_lengths):
        start += norm_dir * length if i % 2 == 0 else orth_dir * width * ((i % 4) - 2)
        points.append(start)
    
    left_edge = [p + orth_dir * width / 2 for p in points]
    right_edge = [p - orth_dir * width / 2 for p in points]
    
    return left_edge, right_edge

def generate_track():
    # Define the track dimensions and segments
    straight_length = 100
    hairpin_radius = 20
    chicane_segment_lengths = [20, 10, 20]
    track_width = 5

    # Generate track segments
    straight_start = np.array([0, 0])
    straight_end = np.array([straight_length, 0])
    left_straight, right_straight = generate_straight(straight_length, track_width, straight_start)
    
    hairpin_center = straight_end + np.array([0, -hairpin_radius])
    left_hairpin, right_hairpin = generate_hairpin(hairpin_radius, track_width, hairpin_center)
    
    chicane_start = left_hairpin[-1]
    left_chicane, right_chicane = generate_chicane(chicane_segment_lengths, track_width, chicane_start)

    # Combine track segments
    left_track = left_straight + left_hairpin[::-1] + left_chicane
    right_track = right_straight + right_hairpin[::-1] + right_chicane

    # move center to (100, 100)
    left_track = left_track + np.array([100, 100])
    right_track = right_track + np.array([100, 100])

    return left_track, right_track

def convert_to_grid_coordinates(cone, map_size_meters, map_resolution):
    x, y = cone
    grid_x = int(x / map_resolution)
    grid_y = int(y / map_resolution)
    return grid_x, grid_y

def create_occupancy_grid(left_cones, right_cones, map_size_meters, map_resolution):
    map_size_cells = int(map_size_meters / map_resolution)
    grid_image = np.full((map_size_cells, map_size_cells), -1, dtype=np.int8)

    max_weight = 100
    min_weight = 50
    weight_decrement = 5
    cones = []

    for cone in np.concatenate([left_cones, right_cones], axis=0):
        left_x, left_y = convert_to_grid_coordinates(cone, map_size_meters, map_resolution)

        cones.append((left_x, left_y))
    cones = np.array([cones], dtype=np.int32)
    cv2.fillPoly(grid_image, cones, color=0)

    return grid_image

def plot_occupancy_grid(grid, map_size_meters, map_resolution):
    plt.imshow(grid, cmap='gray', origin='lower')
    plt.title('Occupancy Grid ({}m x {}m, {}m/pixel)'.format(map_size_meters, map_size_meters, map_resolution))
    plt.xlabel('X Position (m)')
    plt.ylabel('Y Position (m)')
    plt.xticks(np.arange(0, map_size_meters/map_resolution, step=10), np.arange(0, map_size_meters, step=10*map_resolution))
    plt.yticks(np.arange(0, map_size_meters/map_resolution, step=10), np.arange(0, map_size_meters, step=10*map_resolution))
    plt.show()

# Test Parameters
map_size_meters = 200
map_resolution = 0.1  # 1 meter per pixel for simplicity

# Generate and plot test track
left_cones, right_cones = generate_track()
grid = create_occupancy_grid(left_cones, right_cones, map_size_meters, map_resolution)
plot_occupancy_grid(grid, map_size_meters, map_resolution)







